#!/usr/bin/env python
# www.byvac.com
# Contains file transfer utiltiies for use with BvSerial and any other
# Python code
#
# * BMP file sender (bv501 format)
# * BFF font sender (bv501 format)
# * .bas file sender with # directives:
#       #include
#       #comment
#       #send
#       #option echo
#       #option first
#       #option only
# * SV3 Utilities
# loop break

import urllib
import sys, struct, os.path
from time import sleep
from stat import *

ACK = 6
IACK = 62 # inverted SV3 ACK
TOUT = 4
NACK = 21
BNACK = 5 # used by bypic
BLOCKSIZE = 4096
BUFSIZE = 64

# ==============================================================================
# U T I L I T I E S
# ==============================================================================
# ******************************************************************************
# upens a url or file
# ******************************************************************************
def furl_open(fname):
    f = 0
    if 'http:' in fname:
        try:
            f = urllib.urlopen(fname)
        except:
            print "Cant open url:"
            print fname
    else:
        try:
            f = open(fname, "rb")
        except:
            print "Cant open file:",fname
    return f

# ******************************************************************************
# Opens a local file or a url and returns with a list of lines
# ******************************************************************************
def get_lines(fname):
    rv = []
    f = furl_open(fname)
    if f:
        for line in f:
            rv.append(line.rstrip())
            #print "line=", line.rstrip()
    rv.append('') # needs an extra CR at end
    return rv

# ==============================================================================
# I M A G E  C O N V E R S I O N
# ******************************************************************************
# functions to convert font and bmp files for the 1.8" display size
# 128 x 160
# The head shown below is for the BV501 but in the details dictionary there
# is enough information to create any head
# both functions return deatals(dictionary), head(list), body(list)
# ******************************************************************************
# Font output structure          BMP
# (1) ID = chr('s')              chr('B')
# (1) number of 4096 blocks
# (2) spare
# (4) actual OUTPUT file size
# (1) cellH                     height high
# (1) cellW                     height low
# (1) min cellW                 width high
# (1) max cellW                 width low
# (2) nBytes                    not used
# (2) padding
# (60) File name
# Data starts at offset 96  (0x60)


def getUINT(f):
    v = ord(f.read(1))
    v += ord(f.read(1)) << 8
    v += ord(f.read(1)) << 16
    v += ord(f.read(1)) << 32
    return v

def getUSHORT(f):
    v = ord(f.read(1))
    v += ord(f.read(1)) << 8
    return v

# ******************************************************************************
# Expects a BMP file as an input and will return two items:
# details (dictionary) all detals
# head (list)          formatted for bv501 80 bytes
# databody (list)
# Image is not checked for correct format (24 bit) and size that must fit the
# display - can be smaller but no bigger
# ******************************************************************************
def bmpconv1(fdname, config):
    linecount = 0
    print "*1*"
    try:
        lines = get_lines(fdname) #ipf = open(fname, "rb")
    except:
        print "Cant open file:",fdname
        return
    databody =[]
    head = []
    details = {}
    print "*2*", len(lines)
    # The filename to store should not contain the directory
    path, filename = os.path.split(fdname)
    print "*3*"
    details['filename'] = filename
    print "*4*",filename
    #linecount = 2 #ipf.seek(2)
    fileSize = getUINT(lines[2])
    print "*5*"
    details['filesize'] = fileSize
    #linecount = 10 #ipf.seek(10)
    offset = getUINT(lines[10])
    details['offset'] = offset
    details['headsize'] =getUINT(lines[11])
    #ipf.seek(18)
    width = getUINT(lines[18])
    height = getUINT(lines[19])
    details['width'] = width
    details['height'] = height
#    details['planes'] = getUSHORT(ipf)
    bpp = getUSHORT(lines[20])
    details['bitsperpixel'] = bpp
    #ipf.seek(34)
    size = getUSHORT(lines[34])
    details['size'] = size
#    ipf.seek(46)
#    details['colours'] = getUINT(ipf)

    # 80 byte head formmated for bv501
    blocks = (fileSize / BLOCKSIZE) + 1
    head.append(ord('B')) # ID
    head.append(blocks)
    # spare short as file size needs so be on a 4 byte boundary
    head.append(0)
    head.append(0)
    #file size
    head.append(fileSize & 0xff)
    head.append((fileSize >> 8) & 0xff)
    head.append((fileSize >> 16) & 0xff)
    head.append((fileSize >> 24) & 0xff)
    head.append(height % 256)
    head.append(height / 256)
    head.append(width % 256)
    head.append(width / 256)
    head.append(0) # pad to 16 boundary
    head.append(0)
    head.append(0)
    head.append(0)
    for j in filename:
        head.append(ord(j))
    head.append(13) # '\r'

    while len(head) < 96:
        head.append(0) # pad out to 96

    # data
    # This is the data in the format r,g,b. The data is stored upside down
    # with the top of the image being stored last and so all data is read
    # into fdata first and read starts with height at max
    red = 2; green = 1; blue = 0
    if config.order == '1':
        red=0; green=1; blue=2
    #
    print "rgb=",config.order,red,green,blue 
    linecount = offset #ipf.seek(offset)
    fdata = lines[offse:] #fdata = ipf.read() # all data
    bytes_per_row = size / height
    details['bytesperrow'] = bytes_per_row
    bpr = width*3 # actual bytes per row
    rowpad = (bytes_per_row-bpr)
    #print "comp=",rowpad,size,len(fdata)
    ldata = size - 1 - rowpad
    while ldata > 0:
        row = 0
        while row < bpr:
            #print ldata,row,ord(fdata[ldata])/4,ord(fdata[ldata-1])/4,ord(fdata[ldata-2])/4
            databody.append(ord(fdata[ldata-red])/4)
            databody.append(ord(fdata[ldata-green])/4)
            databody.append(ord(fdata[ldata-blue])/4)
            row +=3
            ldata -=3
        ldata -= rowpad
        #if ldata < size-200: break
    #filp data rows
    for row in range(0,size,bpr):
        databody[row:row+bpr]=reversed(databody[row:row+bpr])
    return details, head, databody

# ******************************************************************************
# Expects a BMP file as an input and will return two items:
# details (dictionary) all detals
# head (list)          formatted for bv501 80 bytes
# databody (list)
# Image is not checked for correct format (24 bit) and size that must fit the
# display - can be smaller but no bigger
# ******************************************************************************
def bmpconv(fname, config):
    try:
        ipf = open(fname, "rb")
    except:
        print "Cant open file:",fname
        return
    databody =[]
    head = []
    details = {}
    # The filename to store should not contain the directory
    path, filename = os.path.split(fname)
    details['filename'] = filename
    ipf.seek(2)
    fileSize = getUINT(ipf)
    details['filesize'] = fileSize
    ipf.seek(10)
    offset = getUINT(ipf)
    details['offset'] = offset
    details['headsize'] =getUINT(ipf)
    ipf.seek(18)
    width = getUINT(ipf)
    height = getUINT(ipf)
    details['width'] = width
    details['height'] = height
#    details['planes'] = getUSHORT(ipf)
    bpp = getUSHORT(ipf)
    details['bitsperpixel'] = bpp
    ipf.seek(34)
    size = getUSHORT(ipf)
    details['size'] = size
#    ipf.seek(46)
#    details['colours'] = getUINT(ipf)

    # 80 byte head formmated for bv501
    blocks = (fileSize / BLOCKSIZE) + 1
    head.append(ord('B')) # ID
    head.append(blocks)
    # spare short as file size needs so be on a 4 byte boundary
    head.append(0)
    head.append(0)
    #file size
    head.append(fileSize & 0xff)
    head.append((fileSize >> 8) & 0xff)
    head.append((fileSize >> 16) & 0xff)
    head.append((fileSize >> 24) & 0xff)
    head.append(height % 256)
    head.append(height / 256)
    head.append(width % 256)
    head.append(width / 256)
    head.append(0) # pad to 16 boundary
    head.append(0)
    head.append(0)
    head.append(0)
    for j in filename:
        head.append(ord(j))
    head.append(13) # '\r'

    while len(head) < 96:
        head.append(0) # pad out to 96

    # data
    # This is the data in the format r,g,b. The data is stored upside down
    # with the top of the image being stored last and so all data is read
    # into fdata first and read starts with height at max
    # Order added version 09
    red = 2; green = 1; blue = 0
    if config.order == '1':
        red=0; green=1; blue=2
    #
    ipf.seek(offset)
    fdata = ipf.read() # all data
    bytes_per_row = size / height
    details['bytesperrow'] = bytes_per_row
    bpr = width*3 # actual bytes per row
    rowpad = (bytes_per_row-bpr)
    #print "comp=",rowpad,size,len(fdata)
    ldata = size - 1 - rowpad
    while ldata > 0:
        row = 0
        while row < bpr:
            #print ldata,row,ord(fdata[ldata])/4,ord(fdata[ldata-1])/4,ord(fdata[ldata-2])/4
            databody.append(ord(fdata[ldata-red])/4)
            databody.append(ord(fdata[ldata-green])/4)
            databody.append(ord(fdata[ldata-blue])/4)
            row +=3
            ldata -=3
        ldata -= rowpad
        #if ldata < size-200: break
    #filp data rows
    for row in range(0,size,bpr):
        databody[row:row+bpr]=reversed(databody[row:row+bpr])
    return details, head, databody

# ******************************************************************************
# Converts a bff file to font information, head formatted to bv501
# details (dictionary)
# databody (list)
# ******************************************************************************
def bffconv(fname):
    try:
        fi = open(fname, "rb")
    except:
        print "unable to open ",fname
        return

    databody =[]
    head = []
    details = {}
    # The filename to store should not contain the directory
    path, filename = os.path.split(fname)
    details['filename'] = filename
    header = fi.read(275)
    imageW, imageH, cellW, cellH = struct.unpack('iiii',header[2:18])
    BPP, base = struct.unpack('bb',header[18:20])
    details['imagewidth'] = imageW
    details['imageheight'] = imageH
    details['cellwidth'] = cellW
    details['cellheight'] = cellH
    details['bpp'] = BPP
    details['base'] = base

    #  pre-processing
    pitch = imageW / cellW
    chars = pitch * (imageH / cellH)
    x = cellW * cellH
    size = (x * chars) / 8;
    nBytes = x / 8
    if x % 8: nBytes+=1 # nearest next
    details['pitch'] = pitch
    details['chars'] = chars
    details['size'] = size
    details['nbytes'] = nBytes

    # using width data get max and minimum values
    nmax=0
    nmin=255
    for n in header[20:]:
        if ord(n) > nmax: nmax = ord(n)
        if ord(n) < nmin: nmin = ord(n)

    details['maxwidth'] = nmax
    details['minwidth'] = nmin

    # create head
    fileSize = size  # note file size is not the same as output size
    blocks = (fileSize / BLOCKSIZE) + 1
    head.append(ord('s')) # ID
    head.append(blocks)
    # spare short as file size needs so be on a 4 byte boundary
    head.append(0)
    head.append(0)
    #file size
    head.append(fileSize & 0xff)
    head.append((fileSize >> 8) & 0xff)
    head.append((fileSize >> 16) & 0xff)
    head.append((fileSize >> 24) & 0xff)
    head.append(cellH)
    head.append(cellW-1)
#    head.append(cellW)
    head.append(nmin)
    head.append(nmax)
    head.append(nBytes % 256)
    head.append(nBytes / 256)
    head.append(0) # pad so file starts at 0x10
    head.append(0)
    for j in filename:
        head.append(ord(j))
    head.append(13) # '\r'

    while len(head) < 96:
        head.append(0) # pad out to 96

    #body
    mapStart = 275
    widths = header[20:275]
    limit = min(129,chars+base)
    for cha in range(base,limit):
        byc=0x80
        by=0
        counter=0
        row = (cha - base) / pitch
        x = (cha - base) % pitch
        tp = (row*cellH*imageW*3)+(x*cellW*3)
        databody.append(ord(widths[cha:cha+1]))
        for r in range(0,cellH):
            tp+=3
            for c in range(1,cellW):
                fi.seek(mapStart+tp)
                dat = ord(fi.read(1))
                if dat:
                    by|=byc
                tp+=3
                byc>>=1
                if not byc:
                    databody.append(by & 0xff)
                    byc = 0x80
                    by = 0
                    counter+=1
            tp+= (imageW*3)-(cellW*3)
        counter+=1
        while counter < nBytes:
            databody.append(by & 0xff)
            counter+=1
    return details, head, databody

# ==============================================================================
# T R A N S F E R  T O O L S
# ------- Media upload protocol -----------
# Traget will be waiting for host to send an initial command:
# Command 97 : erase all flash, send ack when done
# Command 98,from,blocks : erase number of blocks ack when done
# Command 99, blocks, filename : recieve file
# Trarget will send ACK to start transfer
# Command 27 : finish
# Error code is sent after a NACK:
# 20 - cant carry out that erase
# 19 - memory full
# 18 - file exists
# 17 - no room for file
#
def error_message(msg):
    if msg == 20:
        print "can't carry out that erase"
    elif msg == 19:
        print "memory full"
    elif msg == 18:
        print "filename already exists"
    elif msg == 17:
        print "no room for file"
    else:
        print "unknown error number: ",msg

# ******************************************************************************
# waits for ack from target with timeout this is the primary handshake method
# This is quicker than setting a larger timout on the serial port
# ******************************************************************************
#def wack(sp):
#    timeout = 100
#    while timeout > 1:
#        timeout -= 1
#        c = sp.read(1)
#        if c == chr(NACK): return NACK
#        if c == chr(ACK): return ACK
#    return NACK
def wack(sp):
    #wiat for ack from target with timeout
    timeout = 100
    error = ''
    while timeout > 0:
        timeout -= 1
        c = sp.read(1)
        if c == chr(ACK): break
        if c == chr(NACK): break
        if c == chr(BNACK): break
        if len(c) > 0: error = error + c
    if timeout < 1: return TOUT, error
    else: return ord(c), error
def wack1(sp,timeout):
    #wiat for ack from target with timeout
    error = ''
    while timeout > 0:
        timeout -= 1
        c = sp.read(1)
        if c == chr(ACK): break
        if c == chr(NACK): break
        if c == chr(BNACK): break
        if len(c) > 0: error = error + c
    if timeout < 1: return TOUT, error
    else: return ord(c), error
# ******************************************************************************
# sends a block of data and waits for ack from target A block begins by
# the host sending the size and then the target receiving that number of bytes.
# When all bytes are sent the target should send an ACK. Returns the ACK
# or NACK from the target Can send a 2 byte or 4 byte length value
# ******************************************************************************
def send_block(sp,data,dlen,timeout):
    length = len(data)
    #print "length=",length
    if dlen == 4:
        sp.write(chr(length >> 24))
        sp.write(chr(length >> 16))
    sp.write(chr(length >> 8)) # send length first
    sp.write(chr(length % 256)) # send length first
    # send data
    for n in data:
        # caters for integer or str types
        if isinstance(data, str):
            sp.write(n)
        else:
            sp.write(chr(n))
    sp.flush() # Wait for output buffer to drain.
    # wait for ACK from target
    rv,error = wack1(sp,timeout)
    #print "rv=",rv," error from ack is ",error
    return  rv

# ******************************************************************************
# ******************************************************************************
def flushinput(sp):
    while len(sp.read(1)) > 0:
        pass

# ******************************************************************************
# Send a filename to BV501 to see ifit exists, see description in bv501.c
# for return values, at the moment:
# -1 full
# -4 corrupt memory
# return < 10000 is exsisting block + 10000
# return > 10000 is next free - 10000
# ******************************************************************************
def file_exists(sp,fname):
    # ByPic command is print atfind("filename")[a]nnn[d][a]ok
    bls = ''
    sp.write("\n")  # gets rid of any residulal 'ok'
    sleep(0.1)
    flushinput(sp)
    sp.write('print atfind("'+fname+'")\n')
    sleep(0.2) # wait for response
    for n in range(0,250):
        d = sp.read(1)
        if len(d):
            #print d,ord(d)
            if ord(d) == 10:  # block # is followed by \n
                while True:
                    c = sp.read(1)
                    if len(c) == 0 or ord(c) == 13: break
                    #print '>>',c,ord(c)
                    bls += c
                break
    #print type(bls),len(bls),bls
    if len(bls) > 0: return int(bls)
    else: return 2000

def test(sp,s):
    sp.write(chr(0)) # send length first
    sp.write(chr(5)) # send length first
    sp.write(chr(1))
    sp.write(chr(2))
    sp.write(chr(3))
    sp.write(chr(4))
    sp.write(chr(5))
    #
    sp.write(chr(0))
    sp.write(chr(0))
    sp.write(chr(0))
    sp.write(chr(0))





# ******************************************************************************
# Media transfer: Sends a file to the target. This comprises of the head
# information followed by the data
# ******************************************************************************
def media_transfer(sp,details,head,body,duplicate):
    # transfer protocol start
    fe = file_exists(sp,details['filename'])
    #print "fe(1)=",fe
    if fe < 0 and fe > -5:
        print "AT memory link error(%s)" % details['filename']
        return -1
    #print "fe=",fe
    if fe < -5000 and duplicate == 'deldup':
        while fe < -5000:
            #print "fe(2)=",fe
            fe += 10000
            adr = fe * BLOCKSIZE
            print "Delete at block ",fe
            sp.write("atput("+str(adr)+",0)\n") # delete duplicate
            fe = file_exists(sp,details['filename']) # get next or free block
    if (fe < -5000 and duplicate == 'dup') or fe >= 10000:
        flushinput(sp)
        if fe < -5000:
            fe = file_exists(sp,"") #get next free block
            if not fe >= 10000:
                print "AT memory link error(%s)" % details['filename']
                return -1
        fe -= 10000
        print "writing","atload("+str(fe)+")\n"
        sp.write("atload("+str(fe)+")\n") # this is the block to load data to
        sleep(0.2)

        for j in range(0,len(head),BUFSIZE):
            print '.',
            res = send_block(sp,head[j:j+BUFSIZE],2,100)
            if res != ACK:
                print "ack error sending head at ",j, "res=",res

        for j in range(0,len(body),BUFSIZE):
            print '.',
            res = send_block(sp,body[j:j+BUFSIZE],2,100)
            if res != ACK:
                print "ack error sending body at ",j

        # terminate file send
        print "\nfile sent"
        sp.write(chr(0)) # terminate
        sp.write(chr(0)) # terminate
        sp.write(chr(0)) # terminate
    else:
        print "File %s existst on Flash, ignoring" % details['filename']
    return 0

# ******************************************************************************
# Media upload depends on command given
# cmd:
# 1 file transfer
# 2 directory transfer
# ******************************************************************************
def media_upload(sp, fdname, duplicate, config):
#    if os.path.isfile(fdname):
    if (fdname.find('.bmp') > 0) or (fdname.find('.bff') > 0):
        # file upload
        filename = fdname
        print "filename: ",filename
        try:
            if filename.find('.bmp') > 0:
                details, head, body = bmpconv(filename, config)
            elif filename.find('.bff') > 0:
                details, head, body = bffconv(filename)
            else:
                print "Expecting .bmp or .bff ",filename
                return
        except:
            print '--- Unable to open ---',  filename
            return
        else:
            media_transfer(sp,details,head,body,duplicate)
            return

    if os.path.exists(fdname):
        # directory upload
        directory = os.listdir(fdname)
        for filename in directory:
            print "File: ",filename,
            transfer = True
            try:
                if filename.find('.bmp') > 0:
                    details, head, body = bmpconv(fdname+'/'+filename,config)
                elif filename.find('.bff') > 0:
                    details, head, body = bffconv(fdname+'/'+filename)
                else:
                    print "Not .bmp or .bff, ignoring ",filename
                    transfer = False
            except:
                sys.stderr.write('--- Unable to open %s ---\n' % filename)
            else:
                if transfer:
                    if(media_transfer(sp,details,head,body,duplicate)): return
        return

    print "No file, directory or unable to open"

# ******************************************************************************
# File upload uses ffload (introduced 2.08 - 26).
# Sends a single file to target, target will create file name
# ffload("filename") will send back size of buffer to use:
# ffload("filename")<CR> echo
# filesize<CR> file size is a string
# ACK indicates file created okay
# ******************************************************************************
def send_file_to_sd(sp, fdname):
    path, filename = os.path.split(fdname) # just filename needed
    try:
        ipf = open(fdname, "rb")
    except:
        print '--- Unable to open ---',  fdname
        return -1
    else:
        # The filename to store should not contain the directory
        # send command to bvnnn
        sp.write('ffload("'+filename+'")\n') #
        sleep(0.2)
        ffload =  sp.read(60)
        # because of echo, command will also be returned, so split needed
        # tmp[0] contains echo
        # tmp[1] contains file size
        # tmp[3] contains ack if file created okay, otherwise nack
        tmp = ffload.split()
        #print "DEBUG tmp=",tmp[0:]
        blen = int(tmp[1])
        print "Buffer length", blen
        if blen == 0:
            print "Buffer space cannot be allocated on host, try resetting host"
            return
        blk = ipf.read(blen)
        if ord(tmp[2]) != ACK:
            print "Error creating file on host: ",filename
            return -1
        while len(blk) == blen:
            print '.',
            res = send_block(sp,blk,2,500)
            if res != ACK:
                print "ack error sending body, error code",res
                return -1
            blk = ipf.read(blen)
        # last bit
        res = send_block(sp,blk,2,500)
        #print "last bit:",blk
        print "\nfile sent"
        sp.write(chr(0)) # terminate
        sp.write(chr(0)) # terminate
        sp.write(chr(0)) # terminate
        return 0

#def walktree(top, callback):
#    '''recursively descend the directory tree rooted at top,
#       calling the callback function for each regular file'''
#
#    for f in os.listdir(top):
#        pathname = os.path.join(top, f)
#        mode = os.stat(pathname)[ST_MODE]
#        if S_ISDIR(mode):
#            # It's a directory, recurse into it
#            path, filename = os.path.split(pathname)
#            print "create directory", filename
#            walktree(pathname, callback)
#        elif S_ISREG(mode):
#            # It's a file, call the callback function
#            callback(pathname)
#        else:
#            # Unknown file type, print a message
#            print 'Skipping %s' % pathname
#def visitfile(fdname):
#    path, filename = os.path.split(fdname)
#    print 'send file', filename

# ******************************************************************************
# File upload uses ffload (introduced 2.08 - 26).
# ffload("filename") will send back size of buffer to use
# cmd:
# 1 file transfer
# 2 directory transfer
# ******************************************************************************
def file_upload(sp, fdname, cmd):
    path, filename = os.path.split(fdname)
    if cmd == 1:
        send_file_to_sd(sp,fdname)
    elif cmd == 2:
        try:
            directory = os.listdir(fdname)
        except:
            print "directory name invalid ",fdname
        else:
            print "not available yet"
            #walktree(fdname, visitfile)
            #for filename in directory:
            #    if os.path.isdir(filename):
            ##        print "Directory :",filename
            #    else:
            #        print filename,



# ******************************************************************************
# sends a text file using the simple ACK protocol
# This is more complex as it handles the #option directives that have include
# files so it needs to be recursive
# ******************************************************************************
def send_file(sp,filename):
    lines = get_lines(filename)
    if len(lines) == 0:
        return
    else:
        print "--- File: %s ---" % filename
        # need the filename without the directory so that it can be displayed
        # using just 80 char wide; also path needed if #include does not
        # specify one
        path, shortName = os.path.split(filename)
        print "path=",path
        print "filename=",shortName

        noDupe = 0 # default duplicate functons
        lineecho = False
        inp = lines[0]
        linecount = 1
        while linecount < len(lines):
            if lineecho: print inp
            if noDupe > 0:
                if 'function' in inp or 'constant' in inp:
                    #print "*** inp=",inp
                    fnline = inp.rsplit()[1]
                    if 'function' in inp:
                        fn = fnline[0:fnline.find('(')]
                    else:   # constants can be defined with flash keyword
                        if fnline == 'flash':
                            fn = inp.rsplit()[2]
                        else:
                            fn = fnline
                    # see if function exists on system already
                    sp.write('#1 '+fn+'\r')
                    res,error = wack(sp)
                    if res == ACK:
                        # skip whole file of just function
                        if noDupe == 2: # rest of file
                            print "keyword %s duplicate, at (%d) skipping file %s" % (fn,linecount,shortName)
                            return False
                        else:
                            print "keyword %s already exists, at (%d) ignoring" % (fn,linecount)                    
                            # constants are either just one line or {}
                            if 'constant' in inp:
                                if '{' in inp:
                                    while linecount < len(lines):
                                        inp = lines[linecount]
                                        linecount += 1
                                        if '}' in inp: break
                                else:
                                    pass # just skip the one line done below
                            else: # must be a function
                                while linecount < len(lines):
                                    inp = lines[linecount]
                                    linecount += 1
                                    #print '--(0) %s -- linecount=%d, len(lines)=%d' % (inp,linecount,len(lines))
                                    if 'endf' in inp: break
                            inp = lines[linecount] # next line after endf
                            linecount += 1
                            if linecount == len(lines): return True
            # pre process functions start with hash, mainly this will set a flag
            # include another file, user has otion of just file name if current 
            # path is to be used
            if inp[0:8] == '#include':
                fn = inp.rsplit('"')[1] # get filename
                print
                if send_file(sp,path+'\\'+fn): return
                inp = lines[linecount]
                linecount += 1
                print "linecount %d inp=%s fn=%s" % (linecount,inp,fn)
                continue
            if inp[0:8] == '#comment':
                # ignore lines until #comment again
                while linecount < len(lines)-1:
                    inp = lines[linecount]
                    linecount += 1
                    if '#comment' in inp:
                        inp = lines[linecount] # next line after #comment
                        linecount += 1
                        break
            if inp[0:5] == '#send':
                # send a command
                cmd = inp.rsplit()[1]
                print "Sending %s as a command" % cmd
                sp.write('#2 '+cmd+'\r')
                sleep(.5) #in case of a long command
                res,error =wack(sp)
                if res == ACK:
                    print "sucessful"
                else:
                    print "command not sent"
                sp.read(1000) # clear any response
                inp = lines[linecount] # next line after #send
                linecount += 1

            if inp[0:7] == '#option':
                options = inp.rsplit()
                # option processing
                if options[1] == 'only':
                    if len(options) == 2:
                        noDupe = 1
                    elif options[2] == 'off':
                        noDupe = 1
                    else:
                        noDupe = 1
                if options[1] == 'first':
                    if len(options) == 2:
                        noDupe = 2
                    elif options[2] == 'off':
                        noDupe = 2
                    else:
                        noDupe = 2
                if options[1] == 'echo':
                    if len(options) == 2:
                        lineecho = True
                    elif options[2] == 'off':
                        lineecho = False
                    else:
                        lineecho = True
                inp = lines[linecount]  # past option
                linecount += 1
                continue
            #print "writing",inp    
            sp.write(inp+'\n')
            if not lineecho:
                if 'function' in inp:
                    print "[%s](%d) function: %s" % (shortName, linecount, inp.rsplit()[1])
                if 'constant' in inp:
                    print "[%s](%d) constant: %s" % (shortName, linecount, inp.rsplit()[1][0:40])
            w,error = wack(sp)
            if w != ACK:
                print
                print "Error in file: %s at line %d" % (shortName,linecount)
                print "Line is: %s" % inp
                print "------ Target reports -------"
                print error
                er = ''
                while True:
                    c = sp.read(1)
                    if len(c) == 0: break
                    er += c
                print er
                print "-----------------------------"
                return True
            inp = lines[linecount]
            linecount += 1
        return False

# ******************************************************************************
# text transfer protocol
# enter with text to send for starting and also full filename including path 
# ******************************************************************************
def tload_transfer(sp,start,filename):
    oldtimeout = sp.timeout
    sp.timeout = .001
    if len(filename) == 0:
        print "No filename given"
        return

    # prep for transfer, this sends commands in turn with
    # a delay between each command
    tmp = start.split()
    for n in tmp:
        sp.write(n+'\n')
        sleep(0.2)
    inp = sp.read(1000) # clear Waiting ...
    if 'Waiting' in inp:
        send_file(sp,filename)
        sp.timeout = oldtimeout
    else:
        print "Problem with start string: ",start
        print "expecting 'Waiting..' from target but not recieved it"

# ==============================================================================
# S V 3
# ******************************************************************************
# Lists sv3 devices on the bus
# ******************************************************************************
def sv3_list(sp):
    oldtimeout = sp.timeout
    sp.timeout = .05
    from sv3B import sv3Bus
    bus = sv3Bus(1,sp)
    res = bus.detect()
    print "-----------------------------------------------------"
    print "-- SV3 Utilities, Devices on the bus:"
    if len(res) > 0:
        count = 0
        for l in res:
            #cAdr = chr(int(l))
            id = bus.ID(l)
            fw = bus.FW(l)
            print "-- [%d] Address %d [%s] ID: %d Firmware: %s" % (count,ord(l), l, id, fw[0]+'.'+fw[1])
            count +=1
        print "-----------------------------------------------------"
    sp.timeout = oldtimeout

# ******************************************************************************
# lists all eeprom locations
# ******************************************************************************
def sv3_eeRead(sp,adr):
    oldtimeout = sp.timeout
    sp.timeout = .05
    from sv3B import sv3Bus,sv3
    bus = sv3Bus(1,sp)
    res = bus.detect()
    if not adr in  res:
        print "Adress %s not found on bus" % address
        print "Addresses are:", res
        return
    device = sv3(bus,adr)
    print "reading EEPROM..."
    elist = device.eeRead(0,256)
    #print elist
    print "    0   1   2   3   4   5   6   7"
    for row in range(0,255,8):
        sys.stdout.write("%02x  " % row)
        for col in range(0,8):
            sys.stdout.write("%02x  " % int(elist[row+col]))
        print
    sp.timeout = oldtimeout

# ******************************************************************************
# Write one byteto specified location
# ******************************************************************************
def sv3_eeWrite(sp,adr,location,byte):
    oldtimeout = sp.timeout
    sp.timeout = .05
    from sv3B import sv3Bus,sv3
    bus = sv3Bus(1,sp)
    res = bus.detect()
    if not adr in  res:
        print "Adress %s not found on bus" % address
        print "Addresses are:", res
        return
    device = sv3(bus,adr)
    print "writing EEPROM..."
    device.eeWrite(location,byte)
    print 'done'
    sp.timeout = oldtimeout


# ******************************************************************************
# sde - serial data exchange
# This is the only function needed for reading and writingto SV3. It is
# similar to spi in that there is always an input and output even if one
# or the other is not required.
# Use:
# The cmd must be the full command including the address and \r at the end
# Example: s = sv3(sp,"aD\r") # get device ID from device with an address 'a'
# Returns:
# 1) NACK on an invalid input or if device does not exist
# 2) The value returned or just ACK
# ******************************************************************************
def sde(com,cmd):
    k = ''
    v = ''
    # send command
    com.flushInput()
    com.write(cmd)
    #print "sending:",cmd # debug
    # return value
    while k != ACK:
        k = com.read(1)
        if k == ACK: # all done
            if len(v) == 0:
                v = ACK
            break
        else:
            if len(k) == 0:
                v = ACK
                break
            else:
                v = v + k
    return v

# ******************************************************************************
# return a list of all devices addresses the serial bus
# Automatically detects and corrects for inversion
# ******************************************************************************
def detect(com):
    rl = []
    for j in range(97,123):
        com.flushInput()
        com.write(chr(j)+"H\r")
        s = com.read(1)
        #if len(s) > 0: print ord(s),j
        # check for inversion and correct
        if s == chr(IACK):
            com.write(chr(j)+"I\r")
            s = chr(ACK)
        if s == chr(ACK):
            rl.append(str(j))
    return rl

# ******************************************************************************
# breaks and endless loop using \ x
# ******************************************************************************
def loop_break(sp):
    sp.setDTR(True)
    sleep(.1)
    sp.setDTR(False)
    sleep(0.1)
    sp.write('\\')
    sleep(0.1)
    tmp = sp.read(1)
    print "tmp=",tmp
    if 'D' in tmp:
        print "Boot loader version 2, reload firmware to get out of loop"
    sp.write('x')

# ******************************************************************************
# ******************************************************************************
def prescan(lookup,inf):
    rv = ''
    con = get_lines(inf) # list ofconstants
    cl=[]
    for n in con:
        cl.append(n.split())
    for n in range(0,len(cl)):
        if len(cl[n]) < 3:
            break
        if cl[n][1] == lookup:
            rv = cl[n][2]
            break
    return rv                

